use std::borrow::Cow;

/// Removes all whitespace from a string if any line breaks are present.
///
/// Example: "foo bar" => "foo bar" but "foo \n bar" => "foobar".
pub(crate) fn remove_whitespace_if_line_break<'s>(input: &'s str) -> Cow<'s, str> {
    let mut state = RemoveWhitespaceIfLineBreakState::Normal;
    for (offset, c) in input.char_indices() {
        match (&mut state, c) {
            (RemoveWhitespaceIfLineBreakState::Normal, '\n') => {
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..offset]);
                state = RemoveWhitespaceIfLineBreakState::HasLineBreak(ret);
            }
            (RemoveWhitespaceIfLineBreakState::Normal, ' ' | '\t') => {
                state = RemoveWhitespaceIfLineBreakState::HasWhitespace(offset);
            }
            (RemoveWhitespaceIfLineBreakState::Normal, _) => {}
            (RemoveWhitespaceIfLineBreakState::HasWhitespace(first_whitespace_offset), '\n') => {
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..*first_whitespace_offset]);
                for c in input[*first_whitespace_offset..offset].chars() {
                    if !c.is_ascii_whitespace() {
                        ret.push(c);
                    }
                }
                state = RemoveWhitespaceIfLineBreakState::HasLineBreak(ret);
            }
            (RemoveWhitespaceIfLineBreakState::HasWhitespace(_), _) => {}
            (RemoveWhitespaceIfLineBreakState::HasLineBreak(_), ' ' | '\t' | '\r' | '\n') => {}
            (RemoveWhitespaceIfLineBreakState::HasLineBreak(ret), _) => {
                ret.push(c);
            }
        }
    }
    match state {
        RemoveWhitespaceIfLineBreakState::Normal => Cow::Borrowed(input),
        RemoveWhitespaceIfLineBreakState::HasWhitespace(_) => Cow::Borrowed(input),
        RemoveWhitespaceIfLineBreakState::HasLineBreak(ret) => Cow::Owned(ret),
    }
}

enum RemoveWhitespaceIfLineBreakState {
    Normal,
    HasWhitespace(usize),
    HasLineBreak(String),
}

/// Removes all line breaks from a string
///
/// Example: "foo bar" => "foo bar" but "foo \n bar" => "foo  bar".
pub(crate) fn remove_line_break<'s>(input: &'s str) -> Cow<'s, str> {
    let mut state = RemoveLineBreakState::Normal;
    for (offset, c) in input.char_indices() {
        match (&mut state, c) {
            (RemoveLineBreakState::Normal, '\n') => {
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..offset]);
                state = RemoveLineBreakState::HasLineBreak(ret);
            }
            (RemoveLineBreakState::Normal, _) => {}
            (RemoveLineBreakState::HasLineBreak(_), '\n') => {}
            (RemoveLineBreakState::HasLineBreak(ret), _) => {
                ret.push(c);
            }
        }
    }
    match state {
        RemoveLineBreakState::Normal => Cow::Borrowed(input),
        RemoveLineBreakState::HasLineBreak(ret) => Cow::Owned(ret),
    }
}

enum RemoveLineBreakState {
    Normal,
    HasLineBreak(String),
}

/// Removes all whitespace from a string if any line breaks are present.
///
/// Example: "foo bar" => "foo bar" but "foo \n bar" => "foobar".
pub(crate) fn coalesce_whitespace_if_line_break<'s>(input: &'s str) -> Cow<'s, str> {
    let mut state = CoalesceWhitespaceIfLineBreakState::Normal;
    for (offset, c) in input.char_indices() {
        match (&mut state, c) {
            (CoalesceWhitespaceIfLineBreakState::Normal, '\n') => {
                // Hit line break without any preceding whitespace
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..offset]);
                ret.push(' ');
                state = CoalesceWhitespaceIfLineBreakState::HasLineBreak {
                    in_whitespace: true,
                    ret,
                };
            }
            (CoalesceWhitespaceIfLineBreakState::Normal, ' ' | '\t') => {
                state = CoalesceWhitespaceIfLineBreakState::HasWhitespace {
                    in_whitespace: true,
                    first_whitespace_offset: offset,
                };
            }
            (CoalesceWhitespaceIfLineBreakState::Normal, _) => {}

            (
                CoalesceWhitespaceIfLineBreakState::HasWhitespace {
                    in_whitespace,
                    first_whitespace_offset,
                },
                '\n',
            ) => {
                // Hit line break with preceding whitespace so we add all the text up to the first whitespace and then process the remaining text coalescing the whitespace.
                let mut ret = String::with_capacity(input.len());
                ret.push_str(&input[..*first_whitespace_offset]);
                let mut sub_loop_in_whitespace = false;
                for c in input[*first_whitespace_offset..offset].chars() {
                    if sub_loop_in_whitespace {
                        if !c.is_ascii_whitespace() {
                            // Preceding character was whitespace but this is not.
                            sub_loop_in_whitespace = false;
                            ret.push(c);
                        }
                        // Do nothing if preceding character was whitespace and this character also is whitespace.
                    } else {
                        if c.is_ascii_whitespace() {
                            // Preceding character was not whitespace but this is.
                            sub_loop_in_whitespace = true;
                            ret.push(' ');
                        } else {
                            // Preceding character was not whitespace and this is not either.
                            ret.push(c);
                        }
                    }
                }
                if !*in_whitespace {
                    // If this line break was the start of whitespace then we need to inject a space character for it.
                    ret.push(' ');
                }
                state = CoalesceWhitespaceIfLineBreakState::HasLineBreak {
                    in_whitespace: true, // This was triggered by a line break which is whitespace.
                    ret,
                };
            }
            (
                CoalesceWhitespaceIfLineBreakState::HasWhitespace {
                    in_whitespace,
                    first_whitespace_offset: _,
                },
                ' ' | '\t',
            ) => {
                *in_whitespace = true;
            }
            (
                CoalesceWhitespaceIfLineBreakState::HasWhitespace {
                    in_whitespace,
                    first_whitespace_offset: _,
                },
                _,
            ) => {
                *in_whitespace = false;
            }
            (
                CoalesceWhitespaceIfLineBreakState::HasLineBreak { in_whitespace, ret },
                ' ' | '\t' | '\r' | '\n',
            ) => {
                if !*in_whitespace {
                    ret.push(' ');
                }
                *in_whitespace = true;
            }
            (CoalesceWhitespaceIfLineBreakState::HasLineBreak { in_whitespace, ret }, _) => {
                *in_whitespace = false;
                ret.push(c);
            }
        }
    }
    match state {
        CoalesceWhitespaceIfLineBreakState::Normal => Cow::Borrowed(input),
        CoalesceWhitespaceIfLineBreakState::HasWhitespace {
            in_whitespace: _,
            first_whitespace_offset: _,
        } => Cow::Borrowed(input),
        CoalesceWhitespaceIfLineBreakState::HasLineBreak {
            in_whitespace: _,
            ret,
        } => Cow::Owned(ret),
    }
}

enum CoalesceWhitespaceIfLineBreakState {
    Normal,
    HasWhitespace {
        in_whitespace: bool,
        first_whitespace_offset: usize,
    },
    HasLineBreak {
        in_whitespace: bool,
        ret: String,
    },
}
